;;; ns-menu-bar.el --- define a default menu bar for NS users

;; Author: Carl Edman
;; Keywords: internal

;; Copyright (C) 1993 Free Software Foundation, Inc.

;; This file is part of GNU Emacs.

;; GNU Emacs is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.

;; GNU Emacs is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with GNU Emacs; see the file COPYING.  If not, write to
;; the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

;;; Code:

(require 'cl)
(require 'ispell)

;; Don't clobber an existing menu-bar keymap, to preserve any menu-bar key
;; definitions made in loaddefs.el.
(or (lookup-key global-map [menu-bar])
    (define-key global-map [menu-bar] (make-sparse-keymap "menu-bar")))

(setq menu-bar-final-items '(buffers windows services hide-app quit))

(define-key global-map [menu-bar quit] '("Quit" . save-buffers-kill-emacs))
(define-key global-map [menu-bar hide-app] '("Hide" . do-hide-emacs))
(define-key global-map [menu-bar services] (cons "Services" (make-sparse-keymap "Services")))
(define-key global-map [menu-bar windows] (make-sparse-keymap "Windows"))
(define-key global-map [menu-bar buffers] (make-sparse-keymap "Buffers"))

(defvar menu-bar-user-menu (make-sparse-keymap "User"))
(define-key global-map [menu-bar user] (cons "User" menu-bar-user-menu))

;; This is just one element of the ediff menu--the first.
(define-key menu-bar-ediff-menu [window]
  '("This Window And Next Window" . compare-windows))

(define-key menu-bar-user-menu [calendar] '("Calendar" . calendar))
(define-key menu-bar-user-menu [rmail] '("Read Mail" . rmail))
(define-key menu-bar-user-menu [gnus] '("Read Net News" . gnus))
(define-key menu-bar-user-menu [bookmark]
  '("Bookmarks" . menu-bar-bookmark-map))
(define-key menu-bar-user-menu [gomoku] '("Gomoku" . gomoku))
(define-key menu-bar-user-menu [hanoi] '("Hanoi" . hanoi))
(define-key menu-bar-user-menu [doktor] '("Doctor" . doctor))

(defvar menu-bar-edit-menu (make-sparse-keymap "Edit"))
(define-key global-map [menu-bar edit] (cons "Edit" menu-bar-edit-menu))

(define-key menu-bar-edit-menu [spell] (cons "Spell" ispell-menu-map))
(define-key menu-bar-edit-menu [fill] '("Fill" . fill-region))
(define-key menu-bar-edit-menu [clear] '("Clear" . delete-region))
(define-key menu-bar-edit-menu [query-replace]
  '("Query Replace" . query-replace))
(define-key menu-bar-edit-menu [re-search-back]
  '("Regexp Search Backwards" . re-search-backward))
(define-key menu-bar-edit-menu [search-back]
  '("Search Backwards" . search-backward))
(define-key menu-bar-edit-menu [re-search-fwd]
  '("Regexp Search" . re-search-forward))
(define-key menu-bar-edit-menu [search-fwd]
  '("Search" . search-forward))
(define-key menu-bar-edit-menu [isearch-back]
  '("Search Backwards" . isearch-repeat-backward))
(define-key menu-bar-edit-menu [isearch-fwd]
  '("Search Forwards" . isearch-repeat-forward))
(define-key menu-bar-edit-menu [isearch] '("Search" . isearch-forward))
(define-key menu-bar-edit-menu [choose-next-paste]
  '("Choose Next Paste" . mouse-menu-choose-yank))
(define-key menu-bar-edit-menu [paste] '("Paste" . yank))
(define-key menu-bar-edit-menu [copy] '("Copy" . kill-ring-save))
(define-key menu-bar-edit-menu [cut] '("Cut" . kill-region))
(define-key menu-bar-edit-menu [undo] '("Undo" . undo))

(defvar menu-bar-file-menu (make-sparse-keymap "File"))
(define-key global-map [menu-bar file] (cons "File" menu-bar-file-menu))

(define-key menu-bar-file-menu [emerge] (cons "Emerge" menu-bar-emerge-menu))
(define-key menu-bar-file-menu [ediff] (cons "Compare" menu-bar-ediff-menu))
(define-key menu-bar-file-menu [epatch] (cons "Apply Patch"
                                              menu-bar-epatch-menu))
(define-key menu-bar-file-menu [kill-buffer]
  '("Kill Buffer" . kill-this-buffer))
(define-key menu-bar-file-menu [print-buffer] '("Print Buffer" . print-buffer))
(define-key menu-bar-file-menu [revert-buffer]
  '("Revert Buffer" . revert-buffer))
(define-key menu-bar-file-menu [write-file]
  '("Save Buffer As..." . write-file))
(define-key menu-bar-file-menu [save-buffer] '("Save Buffer" . save-buffer))
(define-key menu-bar-file-menu [dired] '("Open Directory..." . dired))
(define-key menu-bar-file-menu [open-file]
  '("Open File..." . find-file-other-frame))

(defvar menu-bar-help-menu (make-sparse-keymap "Info"))
(define-key global-map [menu-bar help] (cons "Info" menu-bar-help-menu))

(define-key menu-bar-help-menu [emacs-version]
  '("Show Version" . emacs-version))
(define-key menu-bar-help-menu [ns-bug-report]
  '("Report NeXTstep bug" . ns-submit-bug-report))
(define-key menu-bar-help-menu [report-emacs-bug]
  '("Report General Emacs bug" . report-emacs-bug))
(define-key menu-bar-help-menu [emacs-tutorial]
  '("Emacs Tutorial" . help-with-tutorial))
(define-key menu-bar-help-menu [emacs-news]
  '("Emacs News" . view-emacs-news))
(define-key menu-bar-help-menu [man] '("Man..." . manual-entry))
(define-key menu-bar-help-menu [describe-variable]
  '("Describe Variable..." . describe-variable))
(define-key menu-bar-help-menu [describe-function]
  '("Describe Function..." . describe-function))
(define-key menu-bar-help-menu [describe-key]
  '("Describe Key..." . describe-key))
(define-key menu-bar-help-menu [list-keybindings]
  '("List Keybindings" . describe-bindings))
(define-key menu-bar-help-menu [command-apropos]
  '("Command Apropos..." . command-apropos))
(define-key menu-bar-help-menu [describe-mode]
  '("Describe Mode" . describe-mode))
(define-key menu-bar-help-menu [save-preferences]
  '("Save Preferences" . ns-save-preferences))
(define-key menu-bar-help-menu [info-ns]
  '("Info about Emacs for NS" . info-ns-emacs))
(define-key menu-bar-help-menu [info] '("Info" . info))

(defun kill-this-buffer ()      ; for the menubar
  "Kills the current buffer."
  (interactive)
  (kill-buffer (current-buffer)))

(defun info-ns-emacs ()         ; for the menubar as well
  "Jump to ns-emacs info item."
  (interactive)
  (info "ns-emacs"))

(put 'fill-region 'menu-enable 'mark-active)
(put 'kill-region 'menu-enable 'mark-active)
(put 'kill-ring-save 'menu-enable 'mark-active)
(put 'yank 'menu-enable '(selection-exists-p))
(put 'delete-region 'menu-enable 'mark-active)
(put 'undo 'menu-enable '(if (eq last-command 'undo)
                             pending-undo-list
                           (consp buffer-undo-list)))
(put 'query-replace 'menu-enable (not buffer-read-only))

(defun kill-this-buffer-enabled-p ()
  (let ((count 0)
        (buffers (buffer-list)))
    (while buffers
      (or (string-match "^ " (buffer-name (car buffers)))
          (setq count (1+ count)))
      (setq buffers (cdr buffers)))
    (> count 1)))

(put 'save-buffer 'menu-enable '(buffer-modified-p))
(put 'revert-buffer 'menu-enable
     '(or revert-buffer-function revert-buffer-insert-file-contents-function
          (and (buffer-file-name)
               (or (buffer-modified-p)
                   (not (verify-visited-file-modtime (current-buffer)))))))
;; Permit deleting frame if it would leave a visible or iconified frame.
(put 'delete-frame 'menu-enable
     '(let ((frames (frame-list))
            (count 0))
        (while frames
          (if (cdr (assq 'visibility (frame-parameters (car frames))))
              (setq count (1+ count)))
          (setq frames (cdr frames)))
        (> count 1)))
(put 'kill-this-buffer 'menu-enable '(kill-this-buffer-enabled-p))

(put 'advertised-undo 'menu-enable
     '(and (not (eq t buffer-undo-list))
           (if (eq last-command 'undo)
               (and (boundp 'pending-undo-list)
                    pending-undo-list)
             buffer-undo-list)))

(defvar yank-menu-length 43
  "*Maximum length of an item in the menu for \
\\[mouse-menu-choose-yank].")

(defun mouse-menu-choose-yank (event)
  "Pop up a menu of the kill-ring for selection with the mouse.
The kill-ring-yank-pointer is moved to the selected element.
A subsequent \\[yank] yanks the choice just selected."
  (interactive "e")
  (let* ((count 0)
	 (menu (mapcar (lambda (string)
			 (if (> (length string) yank-menu-length)
			     (setq string (substring string
						     0 yank-menu-length)))
			 (prog1 (cons string count)
			   (setq count (1+ count))))
		       kill-ring))
	 (arg (popup-menu event 
                          (list "Yank Menu"
                                (cons "Choose Next Yank" menu)))))
    ;; A mouse click outside the menu returns nil.
    ;; Avoid a confusing error from passing nil to rotate-yank-pointer.
    ;; XXX should this perhaps do something other than simply return? -rm
    (if arg
	(progn
	  ;; We don't use `rotate-yank-pointer' because we want to move
	  ;; relative to the beginning of kill-ring, not the current
	  ;; position.  Also, that would ask for any new X selection and
	  ;; thus change the list of items the user just chose from, which
	  ;; would be highly confusing.
	  (setq kill-ring-yank-pointer (nthcdr arg kill-ring))
	  (if (interactive-p)
	      (message "The next yank will insert the selected text.")
	    (current-kill 0))))))
(put 'mouse-menu-choose-yank 'menu-enable 'kill-ring)

(defvar complex-buffers-menu-p nil
  "*Non-nil says, offer a choice of actions after you pick a buffer.
This applies to the Buffers menu from the menu bar.")

(defvar buffers-menu-max-size 10
  "*Maximum number of entries which may appear on the Buffers menu.
If this is 10, then only the ten most-recently-selected buffers are shown.
If this is nil, then all buffers are shown.
A large number or nil slows down menu responsiveness.")

(defvar list-buffers-directory nil)

(defun menu-bar-select-buffer ()
  (interactive)
  (switch-to-buffer last-command-event))

(defun menu-bar-update-buffers ()
  (let* ((buffers (buffer-list))
	 buffers-menu)
    ;; If requested, list only the N most recently selected buffers.
    (if (and (integerp buffers-menu-max-size)
             (> buffers-menu-max-size 1))
        (if (> (length buffers) buffers-menu-max-size)
            (setcdr (nthcdr buffers-menu-max-size buffers) nil)))
    
    ;; Make the menu of buffers proper.
    (setq buffers-menu
          (cons "Select Buffer"
                (let ((tail buffers)
                      (maxbuf 0)
                      (maxlen 0)
                      alist
                      head)
                  (while tail
                    (or (eq ?\ (aref (buffer-name (car tail)) 0))
                        (setq maxbuf
                              (max maxbuf
                                   (length (buffer-name (car tail))))))
                    (setq tail (cdr tail)))
                  (setq tail buffers)
                  (while tail
                    (let ((elt (car tail)))
                      (or (eq ?\ (aref (buffer-name elt) 0))
                          (setq alist (cons
                                       (cons
                                        (format
                                         (format "%%%ds  %%s%%s  %%s"
                                                 maxbuf)
                                         (buffer-name elt)
                                         (if (buffer-modified-p elt)
                                             "*" " ")
                                         (save-excursion
                                           (set-buffer elt)
                                           (if buffer-read-only "%" " "))
                                         (abbreviate-file-name
                                          (or (buffer-file-name elt)
                                              (save-excursion
                                                (set-buffer elt)
                                                list-buffers-directory)
                                              "")))
                                        elt)
                                       alist)))
                      (and alist (> (length (car (car alist))) maxlen)
                           (setq maxlen (length (car (car alist))))))
                    (setq tail (cdr tail)))
                  (setq alist (nreverse alist))
                  (nconc (mapcar '(lambda (pair)
                                    ;; This is somewhat risque, to use
                                    ;; the buffer name itself as the event type
                                    ;; to define, but it works.
                                    ;; It would not work to use the buffer
                                    ;; since a buffer as an event has its
                                    ;; own meaning.
                                    (nconc (list (buffer-name (cdr pair))
                                                 (car pair)
                                                 (cons nil nil))
                                           'menu-bar-select-buffer))
                                 alist)
                         (list (cons 'list-buffers
                                     (cons
                                      (concat (make-string (max (- (/ maxlen
                                                                      2)
                                                                   8)
                                                                0) ?\ )
                                              "List All Buffers")
                                      'list-buffers)))))))
    (if buffers-menu
        (setq buffers-menu (cons 'keymap buffers-menu)))
    (define-key global-map [menu-bar buffers]
      (cons "Buffers" (or buffers-menu 'undefined)))))

(add-hook 'menu-bar-update-fab-hook 'menu-bar-update-buffers)

(defun menu-bar-select-frame ()
  (interactive)
  (make-frame-visible last-command-event)
  (raise-frame last-command-event)
  (select-frame last-command-event))

;; ns-arrange functions contributed
;; by Eberhard Mandler <mandler@dbag.ulm.DaimlerBenz.COM>
(defun ns-arrange-all-frames ()
  "Arranges all frames according to topline"
  (interactive)
  (ns-arrange-frames t))

(defun ns-arrange-visible-frames ()
  "Arranges all visible frames according to topline"
  (interactive)
  (ns-arrange-frames nil))

(defun ns-arrange-frames ( vis)
  (let ((frame (next-frame))
	(end-frame (selected-frame))
	(inc-x 20)                      ;relative position of frames
	(inc-y 22)
	(x-pos 100)                     ;start position
	(y-pos 40)
	(done nil))
    (while (not done)                   ;cycle through all frames
      (if (not (or vis (eq (frame-visible-p frame) t)))
	(setq x-pos x-pos); do nothing; true case
	(set-frame-position frame x-pos y-pos)
	(setq x-pos (+ x-pos inc-x))
	(setq y-pos (+ y-pos inc-y))
	(raise-frame frame))
      (select-frame frame)
      (setq frame (next-frame))
      (setq done (equal frame end-frame)))
    (set-frame-position end-frame x-pos y-pos)
    (raise-frame frame)
    (select-frame frame)))

(defun menu-bar-update-frames ()
  (let ((frames (frame-list))
	(frames-menu (make-sparse-keymap "Windows")))
    (define-key frames-menu [popup-color-panel] '("Colors..." . ns-popup-color-panel))
    (define-key frames-menu [popup-font-panel] '("Font Panel..." . ns-popup-font-panel))
    (mapcar
     '(lambda (frame)
        (setcdr frames-menu
                (cons
                 (cons frame (cons (abbreviate-file-name
                                    (cdr (assq 'name (frame-parameters frame))))
                                   'menu-bar-select-frame))
                 (cdr frames-menu))))
     frames)
    (define-key frames-menu [arrange-all-frames]
      '("Arrange All Frames" . ns-arrange-all-frames))
    (define-key frames-menu [arrange-visible-frames]
      '("Arrange Visible Frames" . ns-arrange-visible-frames))
    (define-key frames-menu [delete-frame] '("Delete Frame" . delete-frame))
    (define-key frames-menu [new-frame] '("New Frame" . new-frame))
    (define-key global-map [menu-bar windows] (cons "Windows" frames-menu))))

(add-hook 'menu-bar-update-fab-hook 'menu-bar-update-frames)

(defun menu-bar-update-frames-and-buffers ()
  (if (frame-or-buffer-changed-p)
      (run-hooks 'menu-bar-update-fab-hook)))

(add-hook 'menu-bar-update-hook 'menu-bar-update-frames-and-buffers)

(defconst ns-proto-call-service
  '((arg)
    (interactive "p")
    (let* ((in-string (if mark-active
                          (buffer-substring (region-beginning) (region-end))))
           (out-string (ns-perform-service service in-string)))
      (cond
       ((and out-string (or (not in-string)
                            (not (string= in-string out-string))))
        (if mark-active (delete-region (region-beginning) (region-end)))
        (insert-string out-string)
        (setq deactivate-mark nil))))))
  
(defun ns-define-service (path)
  (let ((mapping [menu-bar services])
        (service (mapconcat 'identity path "/"))
        (name (intern
               (mapconcat '(lambda (s) (if (= s 32) "-" (char-to-string s)))
                          (mapconcat 'identity (cons "ns-service" path) "-") ""))))
    (eval (append (list 'defun name)
                  (subst service 'service ns-proto-call-service)))
    (cond
     ((lookup-key global-map mapping)
      (while (cdr path)
        (setq mapping (vconcat mapping (list (intern (car path)))))
        (if (not (keymapp (lookup-key global-map mapping)))
            (define-key global-map mapping
              (cons (car path) (make-sparse-keymap (car path)))))
        (setq path (cdr path)))
      (setq mapping (vconcat mapping (list (intern (car path)))))
      (define-key global-map mapping (cons (car path) name))))
    name))

(defvar menu-bar-mode nil)

(defun menu-bar-mode (flag)
  "Toggle display of a menu bar on each frame.
This command applies to all frames that exist and frames to be
created in the future.
With a numeric argument, if the argument is negative,
turn off menu bars; otherwise, turn on menu bars."
  (interactive "P")

; ;; Obtain the current setting by looking at default-frame-alist.
; (let ((menu-bar-mode
;        (let ((assq (assq 'menu-bar-lines default-frame-alist)))
;          (if assq (cdr assq) nil))))

  ;; Make menu-bar-mode and default-frame-alist consistent.
  (let ((default (assq 'menu-bar-lines default-frame-alist)))
    (if default
	(setq menu-bar-mode (not (eq (cdr default) 0)))
      (setq default-frame-alist
	    (cons (cons 'menu-bar-lines (if menu-bar-mode 1 0))
		  default-frame-alist))))

  ;; Toggle or set the mode, according to FLAG.
  (setq menu-bar-mode (if (null flag) (not menu-bar-mode)
			(> (prefix-numeric-value flag) 0)))

  ;; Apply it to default-frame-alist.
  (let ((parameter (assq 'menu-bar-lines default-frame-alist)))
    (if (consp parameter)
	(setcdr parameter (if menu-bar-mode 1 0))
      (setq default-frame-alist
	    (cons (cons 'menu-bar-lines (if menu-bar-mode 1 0))
		  default-frame-alist))))
  
  ;; Apply it to existing frames.
  (let ((frames (frame-list)))
    (while frames
      (let ((height (cdr (assq 'height (frame-parameters (car frames))))))
	(modify-frame-parameters (car frames)
				 (list (cons 'menu-bar-lines
					     (if menu-bar-mode 1 0))))
	(modify-frame-parameters (car frames)
				 (list (cons 'height height))))
      (setq frames (cdr frames)))))

(provide 'ns-menu-bar)
(provide 'menu-bar)

;;;; ns-menu-bar.el ends here
